/*
EvolveOp.cpp
Written by Walter O. Krawec

Copyright (c) 2014 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include "EvolveOp.h"

// calculates the probability of finding |actionBasis> \in H_A
// after measuring mixed state rho
// this is simply: tr(M_a * rho) where M_a = I x |a><a|

double EvolveOp::calculatePr(algebra::mat& rho, std::vector <Action>& actions)
{
	if(actions.size() == 1)
	{
		algebra::mat temp;
		M[actions[0].actionSpace][actions[0].actionBasis].multiply(rho, temp);

		return temp.trace().real;	// trace().imag should be 0
	}
	else
		throw std::string("ToDo: calcualtePr, actions.size() > 1");
}

// returns rho := R * M * rho * M * RT / (tr(M * rho))
void EvolveOp::measure(algebra::mat& rho, algebra::mat& M, algebra::mat& R, algebra::mat& RT)
{
	algebra::mat temp;
	M.multiply(rho, temp);
	temp.multiply(M, rho);
	//rho = temp;
	rho.multiplyScalar(algebra::ComplexHelper::invert(temp.trace()));

	R.multiply(rho, temp);
	temp.multiply(RT, rho);
}

// returns rho := \sum_i M[i] \rho M[i]^T
void EvolveOp::measure(algebra::mat& rho, int actionSpace)
{
	algebra::mat temp, temp2, rho2;
	rho2.create(rho.getRows());
	rho2.zero();
	for(unsigned int i=0; i<M[actionSpace].size(); ++i)
	{
		M[actionSpace][i].multiply(rho, temp);
		temp.multiply(M[actionSpace][i], temp2);

		// reset ops
		R[actionSpace][i].multiply(temp2, temp);
		temp.multiply(RT[actionSpace][i], temp2);

		rho2.add(&temp2);
	}

	rho = rho2;
}

void EvolveOp::measure(algebra::mat& rho, std::vector <int>& spacesToMeasure)
{
	if(spacesToMeasure.empty())
	{
		for(int i=0; i<(int)actionDim.size(); ++i)
			measure(rho, i);
	}
	else
	{
		for(unsigned int i=0; i<spacesToMeasure.size(); ++i)
			measure(rho, spacesToMeasure[i]);
	}
}

/*void EvolveOp::evolve(std::vector <Rules>& rules, int maxIterations, int maxPopulation, std::vector <double>* fitnessPoints, std::vector <int>* fitnessAvgNum)
{
	//maxPopulation = 1;
	//maxIterations = 1;
	if(population.empty())
	{
		std::cout << "Population is empty; call prepare() first...\n";
		return;
	}

	std::cout << "Calculating fitness of start population...";
	std::list <CS>::iterator Iter, Iter2;
	int counter=0;
	for(Iter = population.begin(); Iter != population.end(); ++Iter)
	{
		calculateFitness(*Iter, rules);
		if(counter%20)
			std::cout << counter << "\n";
	}
	population.sort();
	std::cout << "done.\n";

	std::cout << "DEBUG: max fitness = " << population.begin()->fitness << "\n";
	std::cout << "DEBUG: min fitness = " << population.rbegin()->fitness << "\n";

	const int CROSSOVER_AMOUNT = 2;
	for(int t=0; t<maxIterations; ++t)
	{
		if(fitnessPoints && fitnessAvgNum && t < (int)fitnessPoints->size())
		{
			(*fitnessAvgNum)[t] ++;
			(*fitnessPoints)[t] += population.begin()->fitness;
		}
#ifdef WIN32
		if(_kbhit()) break;
#endif
		int startingSize = (int)population.size();
		//for(int i=0; i<startingSize / CROSSOVER_AMOUNT + 1; ++i)
		for(int i=0; i<3*startingSize/4+1; ++i)
		{//break;
#ifdef WIN32
			if(_kbhit()) break;
#endif
			//std::cout << i << "/" << startingSize / CROSSOVER_AMOUNT+1 << "\n";
			//int r = rand()%(startingSize / CROSSOVER_AMOUNT + 1);
			int r = rand()%(startingSize/2);
			r = startingSize;
			for(int k=0; k<startingSize/2; ++k)
			{
				int r2 = rand()%startingSize;
				if(r2 < r)
					r = r2;
			}
			Iter = population.begin();
			for(int j=0; j<r; ++j)
			{
				++Iter;
				if(Iter == population.end())
				{
					Iter = population.begin();
					break;
				}
			}
			CS child;
			r = rand()%(startingSize/2);//(startingSize / CROSSOVER_AMOUNT + 1);
			if(rand()%100 < 10)
				r = rand()%startingSize;
			r = startingSize;
			for(int k=0; k<startingSize/2; ++k)
			{
				int r2 = rand()%startingSize;
				if(r2 < r)
					r = r2;
			}
			Iter2 = population.begin();
			for(int j=0; j<r; ++j)
			{
				++Iter2;
				if(Iter2 == population.end())
				{
					Iter2 = population.begin();
					break;
				}
			}

			Iter->crossOver(*Iter2, child);
			calculateFitness(child, rules);
			population.push_back(child);
		}

		std::cout << "mutation...\n";
		Iter = population.begin();
		++Iter;
		for(; Iter != population.end(); ++Iter)
		{//break;
#ifdef WIN32
			if(_kbhit())break;
#endif
			if(rand()%100 < 40)
			{
				Iter->randomEvolve();
				CS q = *Iter;
				calculateFitness(q, rules);
				*Iter = q;
			}
		}

		population.sort();

		//if(population.begin()->fitness < tempPopulation.begin()->fitness)
		{
			for(int i=0; i<startingSize/CROSSOVER_AMOUNT +1; ++i)
			{//break;
#ifdef WIN32
				if(_kbhit()) break;
#endif
				std::cout << i << "/" << startingSize / CROSSOVER_AMOUNT +1 << ": " << population.begin()->fitness << " -> ";
				CS child;
				population.begin()->randomEvolve(child);
				calculateFitness(child, rules);
				std::cout << child.fitness << "\n";
				//if(child.fitness < population.begin()->fitness)
					population.push_back(child);
			}
		}
		

		population.sort();
		int endingSize = (int)population.size();
		while(endingSize > maxPopulation)
		{
			population.pop_back();
			--endingSize;
		}


		for(unsigned int i=0; i<population.begin()->ruleFitness.size(); ++i)
		{
			std::cout << "\tRULE " << i << ": " << population.begin()->ruleFitness[i] << "\n";
		}

		
		//if(population.begin()->fitness <= 0.001) break;
		std::cout << "DEBUG: max fitness = " << population.begin()->fitness << ", " << population.begin()->fitness2 << "\n";
	}

	std::cout << "DEBUG: max fitness = " << population.begin()->fitness << "\n";
	std::cout << "DEBUG: min fitness = " << population.rbegin()->fitness << "\n";

	for(unsigned int i=0; i<U.size(); ++i)
	{
		U[i] = population.begin()->U[i].op;
	}
}*/

void EvolveOp::evolve(std::vector <Rules>& rules, int maxIterations, int maxPopulation, std::vector <double>* fitnessPoints, std::vector <double>* avgFitnessPoints, std::vector <int>* fitnessAvgNum)
{
	//maxPopulation = 1;
	//maxIterations = 1;
	std::list <CS> tempPopulation;
	if(population.empty())
	{
		std::cout << "Population is empty; call prepare() first...\n";
		return;
	}

	std::cout << "Calculating fitness of start population...";
	std::list <CS>::iterator Iter, Iter2;
	int counter=0;
	for(Iter = population.begin(); Iter != population.end(); ++Iter)
	{
		calculateFitness(*Iter, rules);
		//if(counter%20 == 0)
		//	std::cout << counter << "\n";
		++counter;
	}
	population.sort();
	std::cout << "done.\n";

	std::cout << "DEBUG: max fitness = " << population.begin()->fitness << "\n";
	std::cout << "DEBUG: min fitness = " << population.rbegin()->fitness << "\n";

	const int CROSSOVER_AMOUNT = 2;
	for(int t=0; t<maxIterations; ++t)
	{
		tempPopulation.clear();
		if(fitnessPoints && fitnessAvgNum && t < (int)fitnessPoints->size())
		{
			(*fitnessAvgNum)[t] ++;
			(*fitnessPoints)[t] += population.begin()->fitness;

			double avg = 0;
			for(Iter = population.begin(); Iter != population.end(); ++Iter)
				avg += Iter->fitness;
			avg = avg / (double)population.size();
			(*avgFitnessPoints)[t] += avg;
		}
		//if(t > 40) break;
		//if(population.begin()->fitness <= 0.005) break;
#ifdef WIN32
		if(_kbhit()) break;
#endif
		int startingSize = (int)population.size();
		//for(int i=0; i<startingSize / CROSSOVER_AMOUNT + 1; ++i)
		for(int i=0; i<3*startingSize/4+1; ++i)
		{//break;
#ifdef WIN32
			if(_kbhit()) break;
#endif
			//std::cout << i << "/" << startingSize / CROSSOVER_AMOUNT+1 << "\n";
			//int r = rand()%(startingSize / CROSSOVER_AMOUNT + 1);
			int r = rand()%(startingSize/2);
			r = startingSize;
			for(int k=0; k<startingSize/2; ++k)
			{
				int r2 = rand()%startingSize;
				if(r2 < r)
					r = r2;
			}
			Iter = population.begin();
			for(int j=0; j<r; ++j)
			{
				++Iter;
				if(Iter == population.end())
				{
					Iter = population.begin();
					break;
				}
			}
			CS child;
			r = rand()%(startingSize/2);//(startingSize / CROSSOVER_AMOUNT + 1);
			if(rand()%100 < 10)
				r = rand()%startingSize;
			r = startingSize;
			for(int k=0; k<startingSize/2; ++k)
			{
				int r2 = rand()%startingSize;
				if(r2 < r)
					r = r2;
			}
			Iter2 = population.begin();
			for(int j=0; j<r; ++j)
			{
				++Iter2;
				if(Iter2 == population.end())
				{
					Iter2 = population.begin();
					break;
				}
			}

			Iter->crossOver(*Iter2, child);
			calculateFitness(child, rules);
			tempPopulation.push_back(child);
			////population.push_back(child);
		}

		std::cout << "mutation...\n";
		Iter = population.begin();
		++Iter;
		for(; Iter != population.end(); ++Iter)
		{//break;
#ifdef WIN32
			if(_kbhit())break;
#endif
			if(rand()%100 < 40)
			{
				CS child;
				Iter->randomEvolve(child);
				calculateFitness(child, rules);
				tempPopulation.push_back(child);
				//Iter->randomEvolve();
				//CS q = *Iter;
				//calculateFitness(q, rules);
				//*Iter = q;
			}
		}
		//tempPopulation = population;
		tempPopulation.sort();

		int hcPr = 25;
		if(population.begin()->fitness < tempPopulation.begin()->fitness)
		{
			for(int i=0; i<startingSize/CROSSOVER_AMOUNT +1; ++i)
			{//break;
#ifdef WIN32
				if(_kbhit()) break;
#endif
				std::cout << i << "/" << startingSize / CROSSOVER_AMOUNT +1 << ": " << population.begin()->fitness << " -> ";
				CS child;
				population.begin()->randomEvolve(child);
				calculateFitness(child, rules);
				std::cout << child.fitness << "\n";
				if(rand()%100 < hcPr || child.fitness < population.begin()->fitness)
					tempPopulation.push_back(child);
			}
		}
		else
		{
			for(int i=0; i<startingSize/CROSSOVER_AMOUNT +1; ++i)
			{//break;
#ifdef WIN32
				if(_kbhit()) break;
#endif
				//std::cout << i << "/" << startingSize / CROSSOVER_AMOUNT +1 << ": " << tempPopulation.begin()->fitness << " -> ";
				CS child;
				tempPopulation.begin()->randomEvolve(child);
				calculateFitness(child, rules);
				//std::cout << child.fitness << "\n";
				if(rand()%100 < hcPr || child.fitness < tempPopulation.begin()->fitness)
					tempPopulation.push_back(child);
			}
		}

		tempPopulation.sort();
		int endingSize = (int)tempPopulation.size();
		while(endingSize > maxPopulation)
		{
			tempPopulation.pop_back();
			--endingSize;
		}
		Iter = population.begin();
		for(int k=0; k<10; ++k)
		{
			tempPopulation.push_front(*Iter);
			++Iter;
		}

		population = tempPopulation;
		population.sort();

		for(unsigned int i=0; i<population.begin()->ruleFitness.size(); ++i)
		{
			std::cout << "\tRULE " << i << ": " << population.begin()->ruleFitness[i] << "\n";
		}

		std::cout << "DEBUG: max fitness = " << population.begin()->fitness << ", " << population.begin()->fitness2 << "\n";
	}

	std::cout << "DEBUG: max fitness = " << population.begin()->fitness << "\n";
	std::cout << "DEBUG: min fitness = " << population.rbegin()->fitness << "\n";

	for(unsigned int i=0; i<U.size(); ++i)
	{
		U[i] = population.begin()->U[i].op;
	}
}

void EvolveOp::test(std::vector <Rules>& rules)
{
	/*algebra::mat rho;
	rho = initState;
	M[0][0].print(std::cout);
	std::cout << "\ninitState:\n";
	rho.print(std::cout);
	measure(rho, M[0][0], R[0][0], RT[0][0]);
	//measure(rho, 1);
	//measure(rho, 1);
	std::cout << "measured:\n";
	rho.print(std::cout);
	return;
	for(int i=0; i<M.size(); ++i)
	{
		for(int j=0; j<M[i].size(); ++j)
		{
			std::cout << "R[" << i << "][" << j << "]:\n";
			R[i][j].print(std::cout);
		}
	}*/
	CS cs;
	cs.U.resize(U.size());
	for(int i=0; i<(int)U.size(); ++i)
	{
		cs.U[i].op = U[i];
		cs.U[i].op.transpose(cs.U[i].opT);
	}

	calculateFitness(cs, rules, true);
}

void EvolveOp::calculateFitness(CS& cs, std::vector <Rules>& rules, bool debug)
{
	algebra::mat rho;
	rho = initState;

	const double epsilon = 0.1;
	double fitness = 0.0, localFitness;
	double fitness2 = 0.0, localFitness2;
	double smallFit = 0;

	cs.ruleFitness.resize(rules.size());

	// calculate fitness for each rule individually
	// each rule[r] is a vector of SingleRule which are each applied then "measured" (each SingleRule repeated "copies" times)
	// once all SingleRules have been evaluated for rules[r], fitness is calculated for that rule
	// fitness is average of each individual rule[r]
	for(unsigned int r=0; r<rules.size(); ++r)
	{
		if(rules[r].initState.getRows() == 0)
			rho = initState;
		else
			rho = rules[r].initState;
		int counter = 0;
		localFitness = 0.0;
		localFitness2 = 0.0;
		algebra::mat op, opT;
		for(unsigned int k=0; k<rules[r].rules.size(); ++k)
		{
			for(int l=0; l<rules[r].rules[k].copies; ++l)
			{
				for(unsigned int i=0; i<rules[r].rules[k].inputs.inputs.size(); ++i)	// iteration
				{
					switch(rules[r].rules[k].inputs.inputs[i].type)
					{
						case InputOp::EvolvedOp:
							applyOp(rho, cs.U[rules[r].rules[k].inputs.inputs[i].index].op, cs.U[rules[r].rules[k].inputs.inputs[i].index].opT);
							break;
						case InputOp::StaticOp:
							applyOp(rho, V[rules[r].rules[k].inputs.inputs[i].index], VT[rules[r].rules[k].inputs.inputs[i].index]);
							break;
						case InputOp::Measure:
							if(rules[r].rules[k].inputs.inputs[i].subIndex == -1)
								measure(rho, rules[r].rules[k].inputs.inputs[i].index);
							else
								measure(rho, M[rules[r].rules[k].inputs.inputs[i].index][rules[r].rules[k].inputs.inputs[i].subIndex], R[rules[r].rules[k].inputs.inputs[i].index][rules[r].rules[k].inputs.inputs[i].subIndex], RT[rules[r].rules[k].inputs.inputs[i].index][rules[r].rules[k].inputs.inputs[i].subIndex]);
							break;
						case InputOp::RuleOp:
							T[rules[r].rules[k].inputs.inputs[i].index]->evaluate(actionDim, cs.U, V, op, opT);
							applyOp(rho, op, opT);
							break;
						default:
							throw std::string("EvolveOp::calculateFitness(), unknown InputOp type");
					}
					//if(rules[r].rules[k].inputs.inputs[i].type == InputOp::EvolvedOp)
					//	applyOp(rho, cs.U[rules[r].rules[k].inputs.inputs[i].index].op, cs.U[rules[r].rules[k].inputs.inputs[i].index].opT);
					//else
					//	applyOp(rho, V[rules[r].rules[k].inputs.inputs[i].index], VT[rules[r].rules[k].inputs.inputs[i].index]);

					//for(unsigned int j=0; j<rules[r].rules[k].inputs.inputs[i].size(); ++j)	// round
					//	applyOp(rho, cs.U[rules[r].rules[k].inputs.inputs[i][j]].op, cs.U[rules[r].rules[k].inputs.inputs[i][j]].opT);

					//applyOp(rho, V, VT);
				}

				// now calc. prob according to rules:
				double f = 0;
				for(unsigned int i=0; i<rules[r].rules[k].pr.pr.size(); ++i)
				{
					double sum = 0;
					for(unsigned int j=0; j<rules[r].rules[k].pr.pr[i].actions.size(); ++j)
					{
						sum += calculatePr(rho, rules[r].rules[k].pr.pr[i].actions[j]);
					}
					f = f + (sum - rules[r].rules[k].pr.pr[i].p)*(sum - rules[r].rules[k].pr.pr[i].p);
				}

				localFitness2 += (1.0/(f+epsilon));
				//f += epsilon;
				//localFitness += (1.0/f);
				localFitness += f;
				++counter;

				// now measure:
				measure(rho, rules[r].rules[k].spacesToMeasure);
			}
		}

		if(counter > 0)
		{
			localFitness /= (double)counter;
			localFitness2 /= (double)counter;
			//fitness /= (double)counter;
		}
		fitness += localFitness;
		fitness2 += localFitness2;

		cs.ruleFitness[r] = localFitness;
		if(localFitness > smallFit)
			smallFit = localFitness;
	}

	if(rules.size() > 0)
	{
		fitness = fitness / (double)rules.size();
		fitness2 = fitness2 / (double)rules.size();
	}
	cs.fitness = fitness;
	cs.fitness2 = fitness2;

	//cs.fitness = smallFit;
}

// will setup M[i][j] = I x ... x I x |j><j| x I ... x I where the op |j><j| appears in the i'th subspace, for fast lookup later
// other measurement ops (ranging across subspaces) are created as needed and cached (not all may be required)
void EvolveOp::setupMeasurementOps()
{
	int leftSize=1, rightSize=systemSize;
	algebra::mat I1, I2, temp, temp2, Proj;
	M.resize(actionDim.size());
	for(unsigned int i=0; i<actionDim.size(); ++i)
	{
		Proj.create(actionDim[i]);
		Proj.zero();

		// I1 represents I x ... x I "i-1" times (tensor product to the left of |j><j|)
		// I2 represents the tensor product I x ... x I to the right of the |j><j| op
		I1.create(leftSize);
		I1.ident();
		leftSize *= actionDim[i];

		rightSize /= actionDim[i];
		I2.create(rightSize);
		I2.ident();

		M[i].resize(actionDim[i]);
		for(int j=0; j<actionDim[i]; ++j)
		{
			Proj(j,j).real = 1.0;
			I1.tensor(&Proj, &temp);
			temp.tensor(&I2, &temp2);
			M[i][j] = temp2;
			Proj(j,j).real = 0.0;
		}
	}
}

void EvolveOp::setupResetOps()//(bool resetAction, int actionBasis)
{
	R.resize(resetAction.size());
	RT.resize(resetAction.size());
	algebra::mat I1, I2, temp, temp2, Proj;
	int leftSize=1, rightSize=systemSize;
	for(unsigned int i=0; i<resetAction.size(); ++i)
	{
		if(resetAction[i] >= 0)
		{
			Proj.create(actionDim[i]);
			Proj.zero();
			//I.create(memoryDim);
			//I.ident();
			I1.create(leftSize);
			I1.ident();
			leftSize *= actionDim[i];

			rightSize /= actionDim[i];
			I2.create(rightSize);
			I2.ident();
			R[i].resize(actionDim[i]);
			RT[i].resize(actionDim[i]);
			for(int j=0; j<actionDim[i]; ++j)
			{
				Proj(resetAction[i],j).real = 1.0;
				I1.tensor(&Proj, &temp);
				temp.tensor(&I2, &temp2);
				R[i][j] = temp2;
				temp2.transpose(temp);
				RT[i][j] = temp;
				//RT.push_back(temp2);
				Proj(resetAction[i],j).real = 0.0;
			}
		}
		else
		{
			leftSize *= actionDim[i];
			rightSize /= actionDim[i];
			algebra::mat I;
			I.create(systemSize);
			I.ident();
			R[i].resize(actionDim[i]);
			RT[i].resize(actionDim[i]);
			for(int j=0; j<actionDim[i]; ++j)
			{
				R[i][j] = I;
				RT[i][j] = I;
			}
		}
	}
}

bool EvolveOp::saveAsQProg(const std::string& filename)
{
	std::ofstream f;
	f.open(filename.c_str());
	if(!f.is_open())
		return false;

	f << "begin\n";
	for(unsigned int i=0; i<U.size(); ++i)
		f << "\t(new matrix U." << i << " 1 1)\n";

	f << "\n";

	for(unsigned int i=0; i<U.size(); ++i)
	{
		f << "\t(set U." << i << " (matrix " << U[i].getRows() << " " << U[i].getCols() << " ";
		for(int r=0; r<U[i].getRows(); ++r)
		{
			for(int c=0; c<U[i].getCols(); ++c)
			{
				f << "(complex " << U[i](r,c).real << " " << U[i](r,c).imag << ") ";
			}
		}
		f << ") )\n";
	}

	f.close();
	return true;
}

bool EvolveOp::saveAsText(const std::string& filename)
{
	std::ofstream f;
	f.open(filename.c_str());
	if(!f.is_open())
		return false;

	for(unsigned int i=0; i<U.size(); ++i)
	{
		f << U[i].getRows() << "\n" << U[i].getCols() << "\n";
		for(int r=0; r<U[i].getRows(); ++r)
		{
			for(int c=0; c<U[i].getCols(); ++c)
			{
				f << U[i](r,c).real << "\n" << U[i](r,c).imag << "\n";
			}
		}
	}

	f.close();

	return true;
}

bool EvolveOp::saveAsSpin(const std::string& filename)
{
	std::ofstream f;
	f.open(filename.c_str());
	if(!f.is_open())
		return false;

	f << "PUB setupOps\n";
	for(unsigned int i=0; i<U.size(); ++i)
		f << "\tU" << i << " := algebra.create(@heapStart, " << U[i].getRows() << ", " << U[i].getCols() << ")\n";

	f << "\n";
	for(unsigned int i=0; i<U.size(); ++i)
	{
		for(int r=0; r<U[i].getRows(); ++r)
		{
			for(int c=0; c<U[i].getCols(); ++c)
			{
				f << "\talgebra.set(U" << i << ", " << r << ", " << c << ", " << U[i](r,c).real << ", " << U[i](r,c).imag << ")\n";
			}
		}
	}

	f.close();
	return true;
}

void EvolveOp::RuleOp::evaluate(std::vector <int>& A, std::vector <OpData>& U, std::vector <algebra::mat>& V, algebra::mat& opOutput, algebra::mat& opOutputT)
{
	algebra::mat O1, O2, O1T, O2T;
	switch(type)
	{
	case EvolvedOp:
		opOutput = U[arg1].opSmall;
		opOutputT = U[arg1].opSmallT;
		break;
	case StaticOp:
		opOutput = V[arg1];
		opOutput.transpose(opOutputT);
		break;
	case Ident:
		opOutput.create(arg1);
		opOutput.ident();
		opOutputT.create(arg1);
		opOutputT.ident();
		break;
	case Proj:
		opOutput.create(A[arg1]);
		opOutput.zero();
		opOutput(arg2, arg2).real = 1.0;
		opOutput.transpose(opOutputT);
		break;
	case Add:
		T1->evaluate(A, U, V, O1, O1T);
		T2->evaluate(A, U, V, O2, O2T);
		O1.add(&O2, &opOutput);
		opOutput.transpose(opOutputT);
		break;
	case Subtract:
		T1->evaluate(A, U, V, O1, O1T);
		T2->evaluate(A, U, V, O2, O2T);
		O1.subtract(&O2, &opOutput);
		opOutput.transpose(opOutputT);
		break;
	case Mult:
		T1->evaluate(A, U, V, O1, O1T);
		T2->evaluate(A, U, V, O2, O2T);
		O1.multiply(O2, opOutput);
		opOutput.transpose(opOutputT);
		break;
	case Tensor:
		T1->evaluate(A, U, V, O1, O1T);
		T2->evaluate(A, U, V, O2, O2T);
		O1.tensor(&O2, &opOutput);
		opOutput.transpose(opOutputT);
		break;
	}
}