/*
main.cpp
Written by Walter O. Krawec

Copyright (c) 2014 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include <iostream>

#include "EvolveOp.h"
#include "EvolveOpHelper.h"

/** BasicTest Example **/
// Evolve one operator s.t. when applied to (1,1,...1)/sqrt(n) (equal superposition)
// will cause a measurement of |0> to be observed

#include <sstream>
void testAlg()
{
	int n = 8;  // dimension of system
	algebra::Complex16 c;

	// construct the rules:
	std::stringstream str;
	str << "((psi_0):1, (U.0):(1:((0:0))):(all));";

	// convert text rules to data structs for EvolveOp to work with
	EvolveOpHelper e2;
	std::vector <EvolveOp::Rules> ruleList;
	e2.parseRuleString(str.str(), ruleList);

	EvolveOp e;

	// create a default initial state psi = (1,1,...1)/sqrt(n)
	algebra::mat psi;
	std::vector <int> A;
	A.push_back(n);
	int dim = A[0];
	psi.create(dim,1);
	psi.zero();
	c.real = sqrt((double)n);
	c.real = 1.0 / c.real;
	c.imag = 0;
	psi.clear(c);

	e.setup(NULL, NULL, psi, A, NULL, 1);
	e.prepareGA(200);

	e.evolve(ruleList, 80, 200);

	algebra::mat U = e.U[0];	// U is the evolved solution.
}
/** End of Example **/

int main(int argc, char** argv)
{
	testAlg();
	return 0;
}